function [K,Gamma,Info] = lpvsyneng(sys,nmeas,ncont,opt)
% [K,GAM,INFO] = LPVSYNENG(P,NMEAS,NCON,OPT)
%
% K and GAM are the controller and GAM value achieved by the second
% stage LMI (after the backoff).  Solution of the stage 1 LMI is
% in the INFO structure. This code is only for the rate unbounded
% case.
%
% OPT is an (optional) object to specify the options. See LPVSYNOPTIONS.


% Parse Inputs
if nargin==3
    opt = lpvsynOptions;
end
Method = opt.Method;

% Single balancing transformation
% XXX Skip balancing for now
nd = size(sys,2) - ncont;
ne = size(sys,1) - nmeas;
blk = [nd ne; ncont nmeas];
%sysb = lpvbalance(sys,blk);
sysb = sys;

% Orthogonalize general OLIC interconnection
% XXX Non-vectorized for now. Mods to orthog4syn to work on SS arrays
sz = [size(sysb) 1];
nmod = prod(sz(3:end));
[Pdata,TL,TR,FT] = orthog4syn(sysb(:,:,1),nmeas,ncont);
for i=2:nmod
    [Pi,TLi,TRi,FTi] = orthog4syn(sysb(:,:,i),nmeas,ncont);
    Pdata(:,:,i) = Pi;
    TL(:,:,i) = TLi;
    TR(:,:,i) = TRi;
    FT(:,:,i) = FTi;
end

% Get state space data of system P
[a,b,c,d] = ssdata(Pdata);
nX = size(a,1);
ne1 = ne-ncont;
ne2 = ncont;
nd1 = nd-nmeas;
nd2 = nmeas;
d11 = d(1:ne,1:nd,:);
d11dot1 = d(1:ne,1:nd1,:);
d11dot2 = d(1:ne,nd1+1:nd,:);
d111dot = d(1:ne1,1:nd,:);
d112dot = d(ne1+1:ne,1:nd,:);
d1111 = d(1:ne1,1:nd1,:);
d1112 = d(1:ne1,nd1+1:nd,:);
d1121 = d(ne1+1:ne,1:nd1,:);
d1122 = d(ne1+1:ne,nd1+1:nd,:);
d12 = repmat([zeros(ne1,ne2); eye(ne2,ne2)],[1 1 nmod]);
d21 = repmat([zeros(nd2,nd1) eye(nd2,nd2)],[1 1 nmod]);

b11 = b(:,1:nd1,:);
b12 = b(:,nd1+1:nd,:);
b1 = repmat([b11 b12],[1 1 nmod]);
b2 = b(:,nd+1:end,:);
c11 = c(1:ne1,:,:);
c12 = c(ne1+1:ne,:,:);
c1 = repmat([c11;c12],[1 1 nmod]);
c2 = c(ne+1:end,:,:);

for i=1:nmod
    Ahat(:,:,i) = a(:,:,i) - b2(:,:,i)*c12(:,:,i);
    Bhat(:,:,i) = b1(:,:,i) - b2(:,:,i)*d112dot(:,:,i);
    Atil(:,:,i) = a(:,:,i) - b12(:,:,i)*c2(:,:,i);
    Ctil(:,:,i) = c1(:,:,i) - d11dot2(:,:,i)*c2(:,:,i);
end

% Create LMI variables
setlmis([]);
Xdec = lmivar(1,[nX 1]);
Ydec = lmivar(1,[nX 1]);

if isequal(Method,'PoleCon')
    % Hold fixed value of Gamma
    Gamma = opt.Gammaub;
    ginv = 1/Gamma;
    
    %     [ginv,ndec] = lmivar(1,[1 1]);
else
    % Variable representing 1/Gamma
    [ginv,ndec] = lmivar(1,[1 1]);
    
    if isequal(Method,'MaxFeas');
        % LBC is lower bound on X in the coupling constraint
        [LBC,ndec] = lmivar(1,[nX 0]);
    end
end

cnt = 1;
for i=1:nmod
    % Grab state-space data for i-th model
    AhatV = Ahat(:,:,i);
    AtilV = Atil(:,:,i);
    BhatV = Bhat(:,:,i);
    CtilV = Ctil(:,:,i);
    b2V = b2(:,:,i);
    c11V = c11(:,:,i);
    d111dotV = d111dot(:,:,i);
    d11dot1V = d11dot1(:,:,i);
    c2V = c2(:,:,i);
    b11V = b11(:,:,i);
    
    % X Ric LMI
    lmiterm([cnt 1 1 Xdec],AhatV,1,'s');
    lmiterm([cnt 1 2 Xdec],1,c11V');
    lmiterm([cnt 1 1 0],-b2V*b2V');
    lmiterm([cnt 2 2 0],-eye(ne1));
    lmiterm([cnt 3 3 0],-eye(nd));
    if isequal(Method,'PoleCon')
        lmiterm([cnt 1 3 0],ginv*BhatV);
        lmiterm([cnt 2 3 0],ginv*d111dotV);
    else
        lmiterm([cnt 1 3 ginv],1,BhatV);
        lmiterm([cnt 2 3 ginv],1,d111dotV);
    end
    cnt = cnt+1;
    
    % Y Ric LMI
    lmiterm([cnt 1 1 Ydec],1,AtilV,'s');
    lmiterm([cnt 1 2 Ydec],1,b11V);
    lmiterm([cnt 1 1 0],-c2V'*c2V);
    lmiterm([cnt 2 2 0],-eye(nd1));
    lmiterm([cnt 3 3 0],-eye(ne));
    if isequal(Method,'PoleCon')
        lmiterm([cnt 1 3 0],ginv*CtilV');
        lmiterm([cnt 2 3 0],ginv*d11dot1V');
    else
        lmiterm([cnt 1 3 ginv],1,CtilV');
        lmiterm([cnt 2 3 ginv],1,d11dot1V');
    end
    cnt = cnt+1;
    
    % TODO PJS 5/16/2011--Revisit this. Constraints on X/Y do not need
    % to be repeated at each grid point for non-rate bounded problems.
    % There is probably a more general way to avoid repeated constraints.
    if (i==1)
        % X/Y Coupling
        if isequal(Method,'PoleCon')
            lmiterm([-cnt 1 2 0],ginv*eye(nX));
            %             lmiterm([-cnt 1 2 ginv],1,eye(nX));
        else
            lmiterm([-cnt 1 2 ginv],1,eye(nX));
            if isequal(Method,'MaxFeas')
                lmiterm([cnt 1 1 LBC],1,1);
            end
        end
        
        % XXX - These LMIs are redundant if Xlb >0 or Ylb>0
        lmiterm([-cnt 1 1 Xdec],1,1);
        lmiterm([-cnt 2 2 Ydec],1,1);
        cnt = cnt + 1;
        
        % Xlb*I < X < Xub*I
        if opt.Xlb>0
            lmiterm([cnt 1 1 0],opt.Xlb*eye(nX));
            lmiterm([-cnt 1 1 Xdec],1,1);
            cnt = cnt+1;
        end
        if isfinite(opt.Xub)
            lmiterm([-cnt 1 1 0],opt.Xub*eye(nX));
            lmiterm([cnt 1 1 Xdec],1,1);
            cnt = cnt+1;
        end
        
        % Ylb*I < Y < Yub*I
        if opt.Ylb>0
            lmiterm([cnt 1 1 0],opt.Ylb*eye(nX));
            lmiterm([-cnt 1 1 Ydec],1,1);
            cnt = cnt+1;
        end
        if isfinite(opt.Yub)
            lmiterm([-cnt 1 1 0],opt.Yub*eye(nX));
            lmiterm([cnt 1 1 Ydec],1,1);
            cnt = cnt+1;
        end
    end
end

if isequal(Method,'PoleCon')
    % Pole Constraint: eta*[X ginv; ginv Y] > -0.5*(PHI+PHI')
    for i=1:nmod
        % Grab state-space data for i-th model
        AhatV = Ahat(:,:,i);
        AtilV = Atil(:,:,i);
        b11V = b11(:,:,i);
        b12V = b12(:,:,i);
        b2V = b2(:,:,i);
        c11V = c11(:,:,i);
        c12V = c12(:,:,i);
        c2V = c2(:,:,i);
        
        lmiterm([-cnt 1 2 0],eye(nX)/Gamma);
        lmiterm([-cnt 1 1 Xdec],1,1);
        lmiterm([cnt 1 1 Xdec],-0.5*AhatV,1,'s');
        lmiterm([cnt 2 1 Xdec],0.5*(c11V'*c11V)/Gamma,1);
        lmiterm([cnt 1 1 0],b2V*b2V');
        
        lmiterm([-cnt 2 2 Ydec],1,1);
        lmiterm([cnt 2 2 Ydec],1,-0.5*AtilV,'s');
        lmiterm([cnt 2 1 Ydec],1,0.5*(b11V*b11V')/Gamma);
        lmiterm([cnt 2 2 0],c2V'*c2V);
        lmiterm([cnt 1 2 0],-0.5*(b2V*c12V+b12V*c2V)/Gamma);
        
        cnt = cnt+1;
    end
else
    % 1/Gammaub <= 1/Gamma <= 1/Gammalb
    %%gmin = 0.9;	gmax = 500;	gtol = 1;
    %gmin = .1; gmax = 10000; gtol = 1;
    if opt.Gammalb>0
        lmiterm([cnt 1 1 ginv],1,1);
        lmiterm([-cnt 1 1 0],1/opt.Gammalb);
        cnt = cnt +1;
    end
    if isfinite(opt.Gammaub)
        lmiterm([-cnt 1 1 ginv],1,1);
        lmiterm([cnt 1 1 0],1/opt.Gammaub);
        cnt = cnt +1;
    end
end

% TODO PJS 5/25/2011: Add lb/ub constraints on LBC? These constraints are
% in the old code

if ~isequal(Method,'PoleCon')
    % Create objective function:
    % Method = 'MaxFeas':   min -LOB <---> max LOB
    % Method = 'MinGamma' or 'BackOff':  min -1/Gamma <---> min Gamma
    cobj = zeros(ndec,1);
    cobj(end) = -1;
end

% Get LMI Options
% TODO PJS 5/29/2011: Default options for other solvers?
if ~isempty(opt.SolverOptions)
    LMIopt = opt.SolverOptions;
elseif isequal(opt.Solver,'lmilab')
    % Default settings for LMI Lab
    LMIopt = zeros(5,1);
    %LMIopt(1) = 1/(gmax-gmin); % Tol setting in old code
    if isequal(Method,'MaxFeas')
        LMIopt(2) = 40;   % Max # of iters for MaxFeas problem
    elseif 0 % RateBndFlag
        %LMIopt(2) = 600;  % Setting in old LPVOFSYN1
        LMIopt(2) = 250;  % Max # of iters for rate bounded syn
    else
        LMIopt(2) = 150;  % Max # of iters for non-rate bounded syn
    end
    LMIopt(5) = 1;        % Toggle display
else
    LMIopt = [];
end

% Get LMI Initial Condition
if ~isempty(opt.SolverInit)
    x0 = opt.SolverInit;
else
    x0 = [];
end


% Solve LMI
lmisys = getlmis;
FeasFlag = 1;
if isequal(Method,'PoleCon')
    % Pole constraint is a GEVP
    % Currently only implemented using LMILAB/GEVP
    nlfc = nmod;
    [copt,xopt] = gevp(lmisys,nlfc,LMIopt);
    
    % TODO: Add initial conditions
    %[copt,copt] = gevp(lmisys,nlfc,LMIopt,t0,x0);
elseif isequal(opt.Solver,'lmilab')
    [copt,xopt] = mincx(lmisys,cobj,LMIopt,x0);
    if isempty(copt)
        FeasFlag = 0;
    end
elseif isequal(opt.Solver,'sedumi')
    % Convert to Sedumi format
    % TODO PJS 5/29/2011: Currently ignores x0
    [F0,Fi,blk] = lmitrans(lmisys);
    K.s = blk;
    [xdual,xopt,info]=sedumi(Fi,-cobj,F0,K,LMIopt);
    copt = cobj'*xopt;
    if info.pinf==1 || info.dinf==1 || info.numerr~=0
        % TODO PJS 5/29/2011: Also check info.feasratio?
        FeasFlag = 0;
    end
    
else
    % TODO PJS 5/20/2011: Implement other solvers with LMITRANS
    error('Specified solver is currently not available.');
end

% Handle Infeasible LMI Case
% TODO PJS 5/20/2011: What should we return in this case?
if ~FeasFlag
    K = [];
    Gamma = inf;
    Info = struct('xopt',xopt,'copt',copt,'lmisys',lmisys);
    return;
end

% Get optimal X/Y/Gamma variables
if ~isequal(Method,'PoleCon')
    Gamma = 1/dec2mat(lmisys,xopt,ginv);
end

gamsq = Gamma^2;
gamsqi = 1/gamsq;
X = dec2mat(lmisys,xopt,Xdec);
Y = dec2mat(lmisys,xopt,Ydec);

if isequal(Method,'BackOff');
    % Solve Stage 2 LMI: Maximize the Min Eig of X/Y Coupling Constraint
    opt2 = opt;
    opt2.Method = 'MaxFeas';
    opt2.Gammaub = opt.BackOffFactor*Gamma;
    
    % Construct feasible solution from optimal Stage 1 LMI answer
    x0 = [xopt; 0];
    opt2.SolverInit = x0;
    
    % Solve Relaxed LMI
    Info1 = struct('xopt',xopt,'copt',copt,'lmisys',lmisys,'X',X,'Y',Y);
    Gamma1 = Gamma;
    [K,Gamma,Info2] = lpvsyneng(sys,nmeas,ncont,opt2);
    Info = struct('MinGamma',Gamma1,'Stage1Info',Info1,'Stage2Info',Info2);
else
    % Reconstruct Controller
    % TODO PJS 5/20/2011: Put this in a try/catch?
    
    % Compute d(X^-1)/dt = -inv(X)*[ sum_i drho_i/dt* dX/drho_i ]*inv(X)
    Xdot = zeros(nX,nX);
    xinvpart = -X\(Xdot/X);
        
    % Controller Formulas
    K = ss( zeros(ncont,nmeas) );
    for i=1:nmod

        omeg = -d1122(:,:,i) - ...
            d1121(:,:,i)*((gamsq*eye(nd1,nd1) - d1111(:,:,i)'*d1111(:,:,i))\d1111(:,:,i)')*d1112(:,:,i);
        b2omeg = b2(:,:,i)*omeg;
        d12omeg = d12(:,:,i)*omeg;
        
        abar = a(:,:,i) + b2omeg*c2(:,:,i);
        b1bar = b1(:,:,i) + b2omeg*d21(:,:,i);
        c1bar = c1(:,:,i) + d12omeg*c2(:,:,i);
        d11bar = d11(:,:,i) + d12omeg*d21(:,:,i);
        
        dh = inv( eye(ne,ne)- gamsqi*(d11bar*d11bar') );
        dt = inv( eye(nd,nd) - gamsqi*(d11bar'*d11bar) );
        
        F = (-d12(:,:,i)'*dh*d12(:,:,i))\((b2(:,:,i)+b1bar*d11bar'*dh*d12(:,:,i)*gamsqi)'/X + d12(:,:,i)'*dh*c1bar);
        L = -(Y\(c2(:,:,i)+d21(:,:,i)*dt*d11bar'*c1bar*gamsqi)' + b1bar*dt*d21(:,:,i)')/(d21(:,:,i)*dt*d21(:,:,i)');
        
        Af = abar + b2(:,:,i)*F;
        cf = c1bar + d12(:,:,i)*F;
        
        Afx = X\Af;
        left = X\b1bar + cf'*d11bar;
        
        H = -(Afx+Afx'+xinvpart+cf'*cf+gamsqi*left*dt*left');
        
        % Controller State Matrices reconstruction
        % XXX Use shortcourse method
        q = Y - gamsqi*inv(X);
        qiy = q\Y;
        
        m1 = H + F'*(b2(:,:,i)'/X + d12(:,:,i)'*cf);
        %m2 = gamsqi*(gamsq*q*(-qiy*L*d21 - b1bar) + F'*d12'*d11bar)*dt*left';
        m2 = (q*(-qiy*L*d21(:,:,i) - b1bar) + gamsqi*F'*d12(:,:,i)'*d11bar)*dt*left';
        m = m1 + m2;
        
        AC = Af + qiy*L*c2(:,:,i) - gamsqi*(q\m);
        BC = -qiy*L;
        CC = F;
        DC = omeg;
        
        % Undo orthogonalizing transformation
        ortcont = ss(AC,BC,CC,DC);
        K(:,:,i) = lft([TR(:,:,i);eye(ncont)]*ortcont*[TL(:,:,i)' eye(nmeas)],FT(:,:,i));
    end
    
    % Form Info Structure
    Info = struct('xopt',xopt,'copt',copt,'lmisys',lmisys,'X',X,'Y',Y);
end

